#include "stdafx.h"

using namespace std;
using boost::lexical_cast;

struct findFromAddr : public unary_function<TCPSocketBuffer<TimedTCPSocket>, bool> {
	findFromAddr(unsigned long naddr) : naddr_(naddr) {}
	bool operator()(const TCPSocketBuffer<TimedTCPSocket>& s) const
	{
		sockaddr_in sin;
		s.getpeername(reinterpret_cast<sockaddr*>(&sin));
		return sin.sin_addr.s_addr == naddr_;
	}
private:
	unsigned long naddr_;
};

class NodeXMLParser : public XMLParser {
public:
	NodeXMLParser(ChatProcessor* data) : XMLParser()
	{
		XML_SetStartElementHandler(parser, nodeElementStart);
		XML_SetUserData(parser, data);
	}

	virtual ~NodeXMLParser() {}

private:
	static void XMLCALL nodeElementStart(void* data, const char* el, const char** attr);
};

class HandshakeXMLParser : public XMLParser {
public:
	HandshakeXMLParser(ChatProcessor& chatProcessor,
		const TCPSocketBuffer<TimedTCPSocket>& sockbuf)
		: XMLParser(), chatProcessor_(chatProcessor), sockbuf_(sockbuf)
	{
		XML_SetStartElementHandler(parser, elementStart);
		XML_SetUserData(parser, this);
	}
	virtual ~HandshakeXMLParser() {}

private:
	static void XMLCALL elementStart(void* data, const char* el, const char** attr);

	ChatProcessor& chatProcessor_;
	const TCPSocketBuffer<TimedTCPSocket>& sockbuf_;
};

class MessageXMLParser : public XMLParser {
public:
	MessageXMLParser(ChatProcessor& chatProcessor)
		: XMLParser(), ttl(0), chatProcessor_(chatProcessor)
	{
		XML_SetElementHandler(parser, elementStart, elementEnd);
		XML_SetUserData(parser, this);
		XML_SetCharacterDataHandler(parser, charHandler);
	}
	virtual ~MessageXMLParser() {}

private:
	static void XMLCALL elementStart(void* data, const char* el, const char** attr);
	static void XMLCALL elementEnd(void* data, const char* el);
	static void XMLCALL charHandler(void* data, const XML_Char* s, int len);

	ChatProcessor& chatProcessor_;
	string element, guid, name, message, id, sign;
	int ttl;
};

string createGUID(void);
string ipaddr2string(unsigned long addr);

namespace {
	//! ͌Œ肶ȂĊO荞ł悵
	const int maxInConnection = 16;
	const int maxOutConnection = 4;
	const int maxWaitingConnection = 1;
	const char * const chatNodeFileName = "chatnode.txt";
}

TCPServer<TCPSocketBuffer<TimedTCPSocket>, ChatProcessor>* pchatservent;

unsigned int __stdcall chat_thread(void* /*data*/)
{
	unsigned short chat_port;
	try {
		chat_port = lexical_cast<unsigned short>(getIniString("o2on", "chat_port", ""));
	} catch (boost::bad_lexical_cast&) {
		chat_port = 51001;
	}

	// ChatProcessor
	ChatProcessor chatprocessor(chat_port);

	// ChatpserventN
	TCPServer<TCPSocketBuffer<TimedTCPSocket>, ChatProcessor>
		chatservent(chatprocessor, maxInConnection);
	try {
		if (!chatservent.start(chat_port)) {
			throw exception("Chat|[g̃I[vɎs܂\nݒύXčċNĂ");
		}
	} catch (exception& e) {
		// ƂƁc
		MessageBox(NULL, e.what(), "o2on", MB_OK);
		do_thread = false;
		return 0;
	}
	pchatservent = &chatservent;

	while (do_thread) {
		chatservent.process();
	}

	return 0;
}

void ChatProcessor::outboundHandshake(const TCPSocket& s, unsigned short myport, unsigned long yourip)
{
	string handshake = boost::str(boost::format(
		"<o2on><protocol name=\"o2on\" ver=\"0.2\"/>"
		"<client name=\"o2on\" ver=\"0.1\"/>"
		"<nodeinfo myport=\"%1%\" yourip=\"%2%\"/></o2on>") % myport % ipaddr2string(yourip));
	sendP2PMessage(s, "HANDSHAKE", handshake);
}

void ChatProcessor::queueChatMessage(const string& name, const string& message)
{
	Lock lock(cs_msgq);
	msgqueue.push(ChatMessage(name, message, createGUID(), 0, "", ""));
}

string createGUID(void)
{
	// v(w)
	GUID guid;
	CoCreateGuid(&guid);

	return string(boost::str(boost::format("%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X")
		% guid.Data1 % guid.Data2 % guid.Data3 % static_cast<int>(guid.Data4[0])
		% static_cast<int>(guid.Data4[1]) % static_cast<int>(guid.Data4[2])
		% static_cast<int>(guid.Data4[3]) % static_cast<int>(guid.Data4[4])
		% static_cast<int>(guid.Data4[5]) % static_cast<int>(guid.Data4[6])
		% static_cast<int>(guid.Data4[7])));
}

const string ChatMessage::toXmlString() const
{
	int ttl = (ttl_ != -1) ? ttl_ : rand() % 50 + 20;
	return string(str(boost::format("<o2on><name>%1%</name><message>%2%</message>"
		"<guid>%3%</guid><id>%4%</id><sign>%5%</sign><ttl>%6%</ttl></o2on>")
		% name_ % message_ % guid_ % id_ % sign_ % ttl));
}

void ChatMessage::sanitize(string& str)
{
	if (str.empty()) return;
	strreplace(str, "&", "&amp;");
	strreplace(str, "<", "&lt;");
	strreplace(str, ">", "&gt;");
}

void ChatProcessor::sendNodes(const TCPSocket& s)
{
	using namespace boost::lambda;
	string nodesxml = "<o2on>";
	Lock lock = nodes.getCriticalSection();
	vector<NodeInfo>& nodeinfos = nodes.getNodeinfos(lock);

	for (vector<NodeInfo>::iterator it = nodeinfos.begin(), end = nodeinfos.end();
		it != end;
		++it)
	{
		nodesxml += "<node host=\"" + it->getHost() + "\" port=\""
			+ boost::lexical_cast<string>(it->getPort()) + "\"/>";
	}
	nodesxml += "</o2on>";

	sendP2PMessage(s, "NODE", nodesxml);
}

bool ChatProcessor::recvDetected(TCPSocketBuffer<TimedTCPSocket>& sockbuf)
{
	if (sockbuf.getBuffer().length() > 1024 * 1024) {
		return false;
	}
	string& str = sockbuf.getBuffer();

	size_t pos;
	while (pos = str.find("\r\n\r\n"), strfound(pos)) {
		string header(str, 0, pos);
		istringstream iss(header);
		size_t length = 0;
		string method;
		while (iss) {
			string line;
			if (!getline(iss, line)) break;
			size_t pos2 = line.find(": ");
			if (!strfound(pos2)) continue;
			string key(line, 0, pos2);

			string value(line, pos2 + 2);
			if (strfound(pos2 = value.find("\r"))) {
				value.erase(pos2);
			}

			if (key == "Method") {
				method = value;
			} else if (key == "Content-Length") {
				try {
					length = boost::lexical_cast<size_t>(value);
				} catch (boost::bad_lexical_cast&) {}
			}
		}
		if (str.size() < header.size() + 4 + length) return true;	// ܂S͂ĂȂ

		str.erase(0, pos + 4); // Headerobt@

		if (length == 0) return true;

		vector<char> deflateBody(str.begin(), str.begin() + length);
		str.erase(0, length);
		vector<char> body;
		if (!zlib_inflate(deflateBody, body)) {
			// 𓀎s
			return false;
		}

		string strbody(body.begin(), body.end());
		auto_ptr<XMLParser> parser;
		if (method == "MESSAGE") {
			parser.reset(new MessageXMLParser(*this));
		} else if (method == "NODE") {
			parser.reset(new NodeXMLParser(this));
		} else if (method == "HANDSHAKE") {
			parser.reset(new HandshakeXMLParser(*this, sockbuf));
		}

		if (parser.get()) {
			if (!parser->parse(strbody))
				return false;
		}
	}

	return true;
}

void ChatProcessor::dosomething(list<TCPSocketBuffer<TimedTCPSocket> >& sockets)
{
	// ChatMessagẽm[hɓ
	if (!msgqueue.empty()) {
		Lock lock(cs_msgq);
		while (!msgqueue.empty()) {
			for (list<TCPSocketBuffer<TimedTCPSocket> >::iterator it = sockets.begin(),
				end = sockets.end();
				it != end;
				++it)
			{
				const string buf(msgqueue.front().toXmlString());
				sendP2PMessage(it->getSocket(), "MESSAGE", buf);
			}
			msgqueue.pop();
		}
	}

	time_t nowtime;
	time(&nowtime);
	if (nowtime == pretime) {
		// 1secoĂȂꍇ͉Ȃ
		return;
	} else {
		pretime = nowtime;
	}

	// 90bȏ㉽MĂȂؒf
	for (list<TCPSocketBuffer<TimedTCPSocket> >::iterator it = sockets.begin();
		it != sockets.end();
		)
	{
		if (it->getSocket().recvTimeout(90)) {
			it = server_->pop(it);
			disconnected();
		} else {
			++it;
		}
	}

	// 60bȏ㉽MĂȂnode
	for (list<TCPSocketBuffer<TimedTCPSocket> >::iterator it(sockets.begin()), end(sockets.end());
		it != end;
		++it)
	{
		if (it->getSocket().sendTimeout(60)) {
			sendNodes(it->getSocket());
		}
	}

	using namespace boost::lambda;

	// ڑȂꍇOɐڑ
	if (sockets.size() < maxOutConnection && nodes.getSize() > sockets.size()
		&& waitingSockets.size() < maxWaitingConnection)
	{
		Lock lock = nodes.getCriticalSection();
		vector<NodeInfo>& nodeinfos = nodes.getNodeinfos(lock);

		// Dxق sockets.size() + 1 ̃m[h擾
		vector<NodeInfo>::iterator nth = nodeinfos.begin() + sockets.size() + 1;
		nth_element(nodeinfos.begin(), nth, nodeinfos.end(),
			(bind(&NodeInfo::getPriority, _1) > bind(&NodeInfo::getPriority, _2)));

		for (vector<NodeInfo>::iterator it = nodeinfos.begin(); it != nth; ++it) {
			if (find_if(sockets.begin(), sockets.end(), findFromAddr(it->getAddr()))
				!= sockets.end())
			{
				continue;
			}

			TCPSocket new_sock;
			new_sock.nonblocking();
			if (new_sock.connect(it->getAddr(), it->getPort())) {
				SOCKET sock = new_sock.release();
#pragma warning(push, 3)
				FD_SET(sock, &waitingFd);
#pragma warning(pop)
				waitingSockets.push_back(sock);
			}
			it->decPriority();
			break;
		}
	}

	fd_set waiting = waitingFd;
	timeval waitval = {0, 1};
	// ڑłƃnhVFCNserventɓo^
	if (select(FD_SETSIZE, NULL, &waiting, NULL, &waitval) > 0) {
		for (vector<SOCKET>::iterator it = waitingSockets.begin();
			it != waitingSockets.end();
			)
		{
			if (FD_ISSET(*it, &waiting)) {
				TCPSocket new_sock(*it);
				sockaddr_in sin;
				new_sock.getpeername(reinterpret_cast<sockaddr*>(&sin));
				Lock lock = nodes.getCriticalSection();
				vector<NodeInfo>& nodeinfos = nodes.getNodeinfos(lock);
				vector<NodeInfo>::iterator nodesit = find_if(nodeinfos.begin(), nodeinfos.end(),
					bind(&NodeInfo::getAddr, _1) == sin.sin_addr.s_addr);
				if (nodesit != nodeinfos.end()) {
					nodesit->incPriority();
				}

				outboundHandshake(new_sock, chat_port_, sin.sin_addr.s_addr);
				sendNodes(new_sock);	// bIɂŃm[h𓊂
				server_->push(new_sock.release());

#pragma warning(push, 3)
				FD_CLR(*it, &waitingFd);
#pragma warning(pop)
				it = waitingSockets.erase(it);
			} else {
				++it;
			}
		}
	}

	// ڑɎs
	waiting = waitingFd;
	if (select(FD_SETSIZE, NULL, NULL, &waiting, &waitval) > 0) {
		for (vector<SOCKET>::iterator it = waitingSockets.begin();
			it != waitingSockets.end();
			)
		{
			if (FD_ISSET(*it, &waiting)) {
				shutdown(*it, 1);
				closesocket(*it);
#pragma warning(push, 3)
				FD_CLR(*it, &waitingFd);
#pragma warning(pop)
				it = waitingSockets.erase(it);
			} else {
				++it;
			}
		}
	}
}

void ChatProcessor::startProcessor()
{
	// Nodeǂݍ
	ifstream nodeifs(chatNodeFileName);
	if (!nodeifs.fail()) {
		string buf;
		while (getline(nodeifs, buf), !nodeifs.fail()) {
			if (buf.empty())
				continue;

			if (buf[0] == '@') {
				buf.erase(buf.begin());
				buf = decrypt(buf);
				size_t pos = buf.rfind(":");
				if (pos == buf.length() - 1) // |[gw肳ĂȂ
					continue;

				try {
					nodes.pushNodeinfo(NodeInfo(buf.substr(0, pos),
						boost::lexical_cast<unsigned short>(buf.substr(pos + 1))));
				} catch (exception&) {}
			}
		}
	}
}

ChatProcessor::~ChatProcessor()
{
	// Nodeo
	ofstream nodeofs(chatNodeFileName);
	if (!nodeofs.fail()) {
		Lock lock = nodes.getCriticalSection();
		vector<NodeInfo>& nodeinfos = nodes.getNodeinfos(lock);

		for (vector<NodeInfo>::iterator it = nodeinfos.begin(), end = nodeinfos.end();
			it != end;
			++it)
		{
			string buf = it->getHost() + ':' + boost::lexical_cast<string>(it->getPort());
			buf = encrypt(buf);
			nodeofs << '@' << buf << '\n';
		}
	}

	for (vector<SOCKET>::iterator it(waitingSockets.begin()), end(waitingSockets.end());
		it != end;
		++it)
	{
		shutdown(*it, 1);
		closesocket(*it);
	}
}

string ipaddr2string(unsigned long addr)
{
	return boost::str(boost::format("%d.%d.%d.%d") %
		(addr & 0xff) % ((addr >> 8) & 0xff) % ((addr >> 16) & 0xff) % ((addr >> 24) & 0xff));
}

void XMLCALL NodeXMLParser::nodeElementStart(void* data, const char* el, const char** attr)
{
	ChatProcessor* chatProcessor = static_cast<ChatProcessor*>(data);

	string element = el;
	if (element == "node") {
		string host;
		unsigned short port = 0;

		for (int i = 0; attr[i]; i += 2) {
			string name = attr[i];
			if (name == "host") {
				host = attr[i+1];
			} else if (name == "port") {
				try {
					port = boost::lexical_cast<unsigned short>(attr[i+1]);
				} catch (boost::bad_lexical_cast&) {}
			}
		}

		if (!host.empty() && port != 0) {
			try {
				chatProcessor->pushNodeinfo(NodeInfo(host, port));
			} catch (exception&) {}
		}
	}
}

void XMLCALL HandshakeXMLParser::elementStart(void* data, const char* el, const char** attr)
{
	HandshakeXMLParser* parser = static_cast<HandshakeXMLParser*>(data);

	string element = el;
	if (element == "nodeinfo") {
		unsigned short port = 0;

		for (int i = 0; attr[i]; i += 2) {
			string name = attr[i];
			if (name == "myport") {
				try {
					port = boost::lexical_cast<unsigned short>(attr[i+1]);
				} catch (boost::bad_lexical_cast&) {}
			}
		}

		if (port != 0) {
			sockaddr_in sin;
			parser->sockbuf_.getpeername(reinterpret_cast<sockaddr *>(&sin));
			string ipaddr = ipaddr2string(sin.sin_addr.s_addr);
			try {
				parser->chatProcessor_.pushNodeinfo(NodeInfo(ipaddr, port));
			} catch (exception&) {}
		}
	}
}

void XMLCALL MessageXMLParser::elementStart(void* data, const char* el, const char** /*attr*/)
{
	MessageXMLParser* parser = static_cast<MessageXMLParser*>(data);

	parser->element = el;
}

void XMLCALL MessageXMLParser::charHandler(void* data, const XML_Char* s, int len)
{
	MessageXMLParser* parser = static_cast<MessageXMLParser*>(data);

	if (parser->element == "guid") {
		parser->guid.append(s, len);
	} else if (parser->element == "name") {
		parser->name.append(s, len);
	} else if (parser->element == "message") {
		parser->message.append(s, len);
	} else if (parser->element == "ttl") {
		try {
			parser->ttl = lexical_cast<int>(string(s, len));
		} catch (boost::bad_lexical_cast&) {}
	} else if (parser->element == "id") {
		parser->id.append(s, len);
	} else if (parser->element == "sign") {
		parser->sign.append(s, len);
	}
}

void XMLCALL MessageXMLParser::elementEnd(void* data, const char* el)
{
	string element = el;
	if (element != "o2on")
		return;

	MessageXMLParser* parser = static_cast<MessageXMLParser*>(data);
	ChatProcessor& cp = parser->chatProcessor_;

	Lock lock(cp.getMsgqCS());
	deque<string>& processed = cp.getProcessed();
	if (find(processed.begin(), processed.end(), parser->guid) == processed.end()) {
		// GUIDbZ[W͎n߂ēB
		processed.push_back(parser->guid);
		if (processed.size() > 100) {
			processed.pop_front();
		}

		time_t msgtime;
		time(&msgtime);
		Lock lock2(cp.getMsgCS());
		--(parser->ttl);
		ChatMessage chatmsg(parser->name, parser->message, parser->guid, msgtime, parser->id,
			parser->sign, parser->ttl);
		cp.getMessages().push_back(chatmsg);
		if (cp.getMessages().size() > 500) {
			cp.getMessages().pop_front();
		}

		// TTLȂɂ񓊂
		if (parser->ttl > 0) {
			cp.getMsgqueue().push(chatmsg);
		}
	}
}
