// MakeSHA256.cpp: CMakeSHA256 NX̃Cve[V
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MakeSHA256.h"


static void MakeSHA256Helper_c(const void *pBuffer, void *hash, DWORD len);
extern "C" void MakeSHA256Helper_asm(const void *pBuffer, void *hash, DWORD len);

static const DWORD hinit[] = {
	0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19,
};

static const DWORD sha256_k[] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2,
};



////////////////////////////////////////////////////////////////////////////////////////////////////

CMakeSHA256::CMakeSHA256()
{
    Initialize();
}

CMakeSHA256::~CMakeSHA256()
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void CMakeSHA256::Initialize()
{
    CMakeHashAlgorithm::Initialize();
    memcpy(m_pHash, hinit, 32);
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void CMakeSHA256::ComputeHashFinalHelper(void *msg_digest)
{
    BYTE buf[256];
    DWORD len;
    QWORD string_bit_len;

    memcpy(msg_digest, m_pHash, 32);

    len = m_dwRestLen;
    memcpy(buf, m_pRest, len);
    buf[len++] = 0x80;
    if ( len < 56 )
        memset(buf+len, 0, 56-len), len=56;
    else if ( len > 56 )
        memset(buf+len, 0, 120-len), len=120;
    string_bit_len = m_qwSrcLen * 8;
    ReverseEndian64(buf+len, &string_bit_len, 8);
    len += 8;
    ComputeHashHelper(buf, msg_digest, len);

    ReverseEndian32(msg_digest, msg_digest, 32);
}

DWORD CMakeSHA256::ComputeHashHelper(const void *pBuffer, void *msg_digest, DWORD len)
{
#if (USE_ASSEMBLER_SHA256==0)
    MakeSHA256Helper_c(pBuffer, msg_digest, len);
#else
    MakeSHA256Helper_asm(pBuffer, msg_digest, len);
#endif
    return len % 64;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// C

#define CH(x,y,z)       ( (x&(y^z)) ^ z )
#define MAJ(x,y,z)      ( ((x|y)&z) | (x&y) )
#define S0(x)           (ROTATE_R32(x,  2) ^ ROTATE_R32(x, 13) ^ ROTATE_R32(x, 22))
#define S1(x)           (ROTATE_R32(x,  6) ^ ROTATE_R32(x, 11) ^ ROTATE_R32(x, 25))
#define s0(x)           (ROTATE_R32(x,  7) ^ ROTATE_R32(x, 18) ^ (x >> 3))
#define s1(x)           (ROTATE_R32(x, 17) ^ ROTATE_R32(x, 19) ^ (x >> 10))

#define ROUND_sub(a,b,c,d,e,f,g,h,w,k) \
    temp = h + S1(e) + CH(e,f,g) + k + w; \
    d += temp; \
    h = temp + S0(a) + MAJ(a,b,c);

#define ROUND(num) \
        ROUND_sub(a,b,c,d,e,f,g,h,w[num+0],sha256_k[num+0]) \
        ROUND_sub(h,a,b,c,d,e,f,g,w[num+1],sha256_k[num+1]) \
        ROUND_sub(g,h,a,b,c,d,e,f,w[num+2],sha256_k[num+2]) \
        ROUND_sub(f,g,h,a,b,c,d,e,w[num+3],sha256_k[num+3]) \
        ROUND_sub(e,f,g,h,a,b,c,d,w[num+4],sha256_k[num+4]) \
        ROUND_sub(d,e,f,g,h,a,b,c,w[num+5],sha256_k[num+5]) \
        ROUND_sub(c,d,e,f,g,h,a,b,w[num+6],sha256_k[num+6]) \
        ROUND_sub(b,c,d,e,f,g,h,a,w[num+7],sha256_k[num+7])


void MakeSHA256Helper_c(const void *pBuffer, void *pHash, DWORD len)
{
    const BYTE *pBuf = (const BYTE*)pBuffer;
    DWORD *hash = (DWORD*)pHash;
    DWORD w[64];
    DWORD a,b,c,d,e,f,g,h,temp;
    DWORD i, num=len/64;

    a=hash[0]; b=hash[1]; c=hash[2]; d=hash[3];
    e=hash[4]; f=hash[5]; g=hash[6]; h=hash[7];

    while ( num > 0 ) {
        ReverseEndian32(w, pBuf, 64);

	    for(i=16; i<64; i++)
	        w[i] = s1(w[i-2]) + w[i-7] + s0(w[i-15]) + w[i-16];

        ROUND( 0)  ROUND( 8)  ROUND(16)  ROUND(24)
        ROUND(32)  ROUND(40)  ROUND(48)  ROUND(56)

        a+=hash[0];  b+=hash[1];  c+=hash[2];  d+=hash[3];
        e+=hash[4];  f+=hash[5];  g+=hash[6];  h+=hash[7];
        hash[0]=a;   hash[1]=b;   hash[2]=c;   hash[3]=d;
        hash[4]=e;   hash[5]=f;   hash[6]=g;   hash[7]=h;
        pBuf+=64;    num--;
    }
}



////////////////////////////////////////////////////////////////////////////////////////////////////
// test

BOOL CMakeSHA256::Test()
{
    static const char *q1 = "abc";
    static const char *q2 = "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq";

    static const char *a0 = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855";
    static const char *a1 = "ba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff61f20015ad";
    static const char *a2 = "248d6a61d20638b8e5c026930c3e6039a33ce45964ff2167f6ecedd419db06c1";
    static const char *a3 = "cdc76e5c9914fb9281a1c7e284d73e67f1809a48a497200e046d39ccc7112cd0";


    Initialize();
    if ( CmpHashValue(a0) ) {
        ASSERT(0);
        return FALSE;
    }

    Initialize();
    ComputeHash(q1, strlen(q1));
    if ( CmpHashValue(a1) ) {
        ASSERT(0);
        return FALSE;
    }

    Initialize();
    ComputeHash(q2, strlen(q2));
    if ( CmpHashValue(a2) ) {
        ASSERT(0);
        return FALSE;
    }

    Initialize();
    BYTE *pQ3=new BYTE[1000000];
    memset(pQ3, 'a', 1000000);
    ComputeHash(pQ3, 1000000);
    delete [] pQ3;
    if ( CmpHashValue(a3) ) {
        ASSERT(0);
        return FALSE;
    }

    return TRUE;
}

